import {
  getInputBuffer,
  setOutputJSON,
  log,
  getLocation,
  getTagValue,
  getSourceValue,
  getTimestamp,
  getUserdata,
} from "orbit-sdk-assemblyscript";
import { JSONDecoder, JSONHandler } from "./lib/decoder";
import { JSONEncoder } from "./lib/encoder";

/**
 * process uplink (device -> SORACOM) message
 */
export function uplink(): i32 {
  const data = new Data();

  // create a decoder to handle Data class
  const decoder = new JSONDecoder<Data>(data);
  // pass input buffer to the decoder and deserialize
  decoder.deserialize(getInputBuffer());

  // create a new JSONEncoder, based on https://github.com/nearprotocol/assemblyscript-json
  const encoder = new JSONEncoder();
  encoder.setString("name", data.name);

  // get location information
  const location = getLocation();
  encoder.setFloat("lat", location.lat);
  encoder.setFloat("lon", location.lon);

  // get tag value with its name
  const tagname = getTagValue("name");
  encoder.setString("tagname", tagname);
  const tagorg = getTagValue("org");
  encoder.setString("tagorg", tagorg);

  // get source value with its name
  const resourceType = getSourceValue("resourceType");
  encoder.setString("resourceType", resourceType);
  const resourceId = getSourceValue("resourceId");
  encoder.setString("resourceId", resourceId);

  encoder.setInteger("timestamp", getTimestamp());

  // add additional calculation based on device name
  if (data.name == "sorao") {
    encoder.setFloat("value", data.value / 100);
  } else if (data.name == "sorako") {
    encoder.setFloat("value", data.value * 10);
  } else {
    encoder.setInteger("value", -1);
  }

  encoder.setString("userdata", getUserdata());

  log("Hello Orbit!");
  // set output JSON
  setOutputJSON(encoder.toString());

  // return user defined result code for success
  return 0;
}

/**
 * Class representing a simple data.
 *
 * This class extends JSONHandler, event based JSON decoder for AssemblyScript based on https://github.com/nearprotocol/assemblyscript-json.
 * When JSONDecoder receives string or float, relevant event handler (setXxx) will be called with property name and value as parameters.
 */
class Data extends JSONHandler {
  public name: string = "";
  public value: f64 = 0.0;

  /**
   * Set property value if the name is "name"
   * @param name property name
   * @param value property value
   */
  setString(name: string, value: string): void {
    if (name == "name") {
      this.name = value;
    }
  }

  /**
   * Set property value if the name is "value"
   * @param name property name
   * @param value property value
   */
  setFloat(name: string, value: f64): void {
    if (name == "value") {
      this.value = value;
    }
  }
}
